;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :cm)

(defun header (&rest strings)
  (let ((newline (string #\newline)))
    (apply #'concatenate 'string (loop for s in strings 
				       collect s into l
                                       collect newline into l 
                                       finally (return l)))))

(defun fheader (file &rest strings)
  (open-event-stream file ':initialize :header (apply #'header strings)))


;; helper for input file specification. takes either a single string
;; with files names delimited by "," or a list of file names.
;; directories and extensions are sticky, ie: "~/foo, bar"
;; finds "/user/hkt/foo.snd" and "/user/hkt/bar.snd"

(defun files (string &rest others)
  (let (list file)
    (if others (setf list (cons string others))
      (setf list (sequence-to-list string :read nil)))
    (setf file (pop list))
    (cons file
          (loop for f in list
                collect (namestring (setf f (merge-pathnames f file)))
                do (setf file f)))))


(defun play-file (file &rest pairs)
  (when pairs (setf pairs (canonicalize-pairs pairs)))
  (let ((syntax (pair-value ':syntax pairs
                            (or (pathname-syntax file) *syntax*))))
    (unless syntax (error "Can't determine syntax for playing ~S." file))
    (setf file (merge-pathnames file (slot-value syntax 'pathname)))
    (if (probe-file file)
        (apply #'play-using-syntax syntax file pairs)
      (error "File \"~A\" not found.~&" (namestring file)))))

;;;
;;; load-file load a file accoring to a syntax. if lisp, a compile option
;;; is also avaiable.
;;;

(defun load-file (file &rest pairs)
  (setf pairs (canonicalize-pairs pairs nil t :keyword))
  (unless (pathname-type file)
    (let ((types (getf pairs ':file-types ':unspecified)))
      (unless (eq types ':unspecified)
        (remf pairs ':file-types)
        (setf file
              (loop for ext in types
                    for fil = (probe-file 
                               (make-pathname :defaults file :type ext))
                    when fil return fil
                    finally (return file))))))
  (apply #'load file pairs))

#|
(defun load-file (file &rest pairs)
  (let ((syntax (or (pair-value 'syntax pairs)
                    (pair-value ':syntax pairs)
                    (pathname-syntax file))))
    (if syntax
        (apply #'load-using-syntax syntax file pairs)
      (let ((type (or (pathname-type file) "")))
        (cond ((string-equal type "tl")
               (apply #'script-file file pairs))
              ((string-equal type +cm-type+)
               (apply #'cload file :compile nil pairs))
              (t
               (apply #'cload file pairs)))))))

(defvar *cload-source-types*
  (list +cm-type+ "stella"  +source-type+ "tl" "ins" "cm" "clm" "cmn" ))

(defun cload (file &rest pairs)
  (setf file (namestring file))
  (let ((compile nil)
        (type (pathname-type file))
        (verbose t) source loadfile)
    (dopairs (n v pairs)
      (case n 
         ((compile :compile) (setf compile v))
         ((verbose :verbose) (setf verbose v))))
    (if type
        (progn (setf source (probe-file file))
               (when (and (eq compile ':check)
                          (string-equal type +binary-type+))
                 (setf compile nil)))
      (setf source
            (loop for type in *cload-source-types*
                  when (probe-file (make-pathname :defaults file :type type))
                  return it)))
    (unless source
      (warn-user "File \"~A\" not found." file)
      (return-from cload nil))
    (setf loadfile
      (if (not compile) source
        (make-pathname :defaults source :type +binary-type+)))
    (when compile
      (setf source (truename source))
      (when (or (not (probe-file loadfile))
                (or (eq compile t) (eq compile ':force))
                (< (file-write-date loadfile)
                   (file-write-date source)))
        #+(or cltl2 lispworks clisp)
	(compile-file source :output-file loadfile
		      :verbose verbose)
        #-(or cltl2 lispworks clisp)
	(compile-file source :output-file loadfile)))
     (setf loadfile (truename loadfile))
     (load loadfile :verbose verbose)
     loadfile))
|#

;;;
;;; import-file imports a file according to a syntax.
;;;

(defun import-file (file &rest pairs)
  (when pairs (setf pairs (canonicalize-pairs pairs)))
  (let ((syntax (pair-value 'syntax pairs
                            (or (pathname-syntax file) *syntax*))))
    (apply #'import-using-syntax syntax file pairs)))

