;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :cm)

(defsyntax m4c
  :pathname (pathname "test.m4c")
  :stream-types '((m4c-score-file "sc")))

(defparameter *m4c-orchestra* "/usr/CMP/m4c/m4c/classorch/m4c.class" )
(defparameter *m4c-args* "-NH -e" )
(defparameter *m4c-player* 
  (namestring (merge-pathnames "playsnd" *cm-site-directory*)))

(defobject m4c-score-file (event-file)
  ((syntax :initform (find-syntax ':m4c))
   (orchestra :initform  *m4c-orchestra*
              :initarg orchestra :initarg :orchestra)
   (channels :initform 1 :initarg channels :initarg :channels)
   (srate :initform 22050 :initarg srate :initarg :srate)
   (output :initarg output :initarg :output)
   (play :initform t :initarg play :initarg :play)
   (listing :initarg listing :initarg :listing)
   (args :initform *m4c-args* :initarg args :initarg :args))
  (:parameters start end timescale
               orchestra channels srate output listing args))

(defmethod initialize-stream-for-processing ((stream m4c-score-file))
  (let ((file (slot-value stream 'stream)))
    (format file "/* Common Music output of ~A */~%~%" (date-string))))

(defmethod deinitialize-stream-for-processing ((stream m4c-score-file))
  (write-line "E" (slot-value stream 'stream))) 

#+unix
(defmethod play-using-syntax ((syntax m4c) file &rest arglist)
  (let* ((stream (open-event-stream file))
         (orch (or (pair-value 'orchestra arglist)
                   (slot-value stream 'orchestra)))
         (ofil (or (pair-value 'output arglist)
                   (slot-value-or-default
                    stream 'output
                    (namestring 
                     (make-pathname :type "snd" :defaults file)))))
         (lfil (or (pair-value 'listing arglist)
                   (slot-value-or-default
                    stream 'listing
                    (namestring 
                     (make-pathname :type "list" :defaults file)))))
         (args (or (pair-value 'args arglist)
                   (slot-value stream 'args)))
         (chan (or (pair-value 'args arglist)
                   (slot-value stream 'channels)))
         (rate (or (pair-value 'srate arglist)
                   (slot-value stream 'srate)))
         (play (if (or (pair-value 'play arglist) 
                       (slot-value stream 'play))
                 (format nil "~A ~A" *m4c-player* ofil)))
         (cmd (format nil "~A ~A -c~A -s~A ~A ~A >& ~A~@[; ~A~] &"
                      orch args chan rate ofil file lfil play)))
     (tell-user "Playing: ~A~%" cmd)
    (shell cmd)))

#-unix
(defmethod play-using-syntax ((syntax rt) file &rest pairs)
  (declare (ignore file pairs))
  (warn-user "Sorry, can't play M4C scores on a non-Unix machine."))

(defclass m4c-note (note)
  ((ins :initarg ins :initarg :ins)
   (dur :initarg dur :initarg :dur)
   (lfrac :initarg lfrac :initarg :lfrac)  ; optional speaker location
   ))

;;;
;;; m4c instruments typically expect pitch to be encoded as a float:
;;; octave.interval, where octave is the keyboard octave with 4 added
;;; to it. this notation was invented at Princeton, presumably after
;;; the folks there had drunk a few too many beers.
;;;

(defun oct.pitch (freq)
  ;; convert standard notations to princeton's octave.interval notation
  (multiple-value-bind (o i) (octave-and-interval freq *standard-scale*)
    (+ 3 o (* .01 i))))

(defmethod print-object ((object m4c-note) stream)
  (format stream "#<~A ~A ~A"
          (slot-value-or-default object 'ins +slot-unset+)
          (slot-value-or-default object 'rhythm +slot-unset+)
          (slot-value-or-default object 'dur +slot-unset+))
  (dolist (p (object-parameters object))
    (write-char #\Space stream)
    (princ (slot-value-or-default object (parameter-name p) +slot-unset+)
           stream))
  (write-char #\> stream))

(defmacro fformat (f d)
  (let ((v (gensym)) )
    `(let ((,v ,f))
       (if (floatp ,v)
         (/ (the fixnum (round (the float ,v)
                               ,(expt 10.0 (- d)))) 
            ,(expt 10.0 d))
         ,v))))

(defmethod initialize-instance :after ((object m4c-note) &rest args)
  (declare (ignore args))
  (unless (slot-boundp object 'ins)
    (setf (slot-value object 'ins)
          (string-capitalize
           (symbol-name (class-name (class-of object)))))))

(defmethod write-event ((object m4c-note) (stream m4c-score-file))
  (let ((file (slot-value stream 'stream))
        (delim " {")
        name value)
    (format file "I(~A, ~S, ~S)" (slot-value object 'ins)
            (fformat (slot-value object 'time) 3)
            (fformat  (slot-value object 'dur) 3))
    (dolist (param (object-parameters object))
      (setf name (parameter-name param))
      (setf value (if (eq name 'lfrac)
                    (slot-value-or-default object name nil)
                    (slot-value object (parameter-name param))))
      (when value
        (write-string delim file)
        (princ (fformat value 3) file)
        (setf delim ", ")))
    (write-line "}" file)))

;;;
;;; m4c.class templates.  Class names and parameters follow the
;;; M4C tutorial with the following exceptions:
;;; Pluck:
;;;   1 Renamed Pl because the Music Kit already claims Pluck.
;;;   2 PITCH parameter is renamed OCT.PITCH
;;; Glide, Gl1-10:
;;;   START_PITCH is renamed oct.pitch
;;; GFIns:
;;;   START_PITCH is renamed oct.pitch
;;;

(defobject Pl (m4c-note)
  ((ins :initform "Pluck")
   oct.pitch amp db_decay)
  (:parameters oct.pitch amp db_decay lfrac))

(defobject Pladsr (m4c-note)
  ((ins :initform "Pladsr")
   oct.pitch amp db_decay attack decay sus release)
  (:parameters oct.pitch amp db_decay attack decay sus release lfrac))

(defobject Glide (m4c-note)
  ((ins :initform "Glide")
   oct.pitch amp attack decay pitch_change)
  (:parameters oct.pitch amp attack decay pitch_change lfrac))

(defobject gl1 (glide) ((ins :initform "GL1")))
(defobject gl2 (glide) ((ins :initform "GL2")))
(defobject gl3 (glide) ((ins :initform "GL3")))
(defobject gl4 (glide) ((ins :initform "GL4")))
(defobject gl5 (glide) ((ins :initform "GL5")))
(defobject gl6 (glide) ((ins :initform "GL6")))
(defobject gl7 (glide) ((ins :initform "GL7")))
(defobject gl8 (glide) ((ins :initform "GL8")))
(defobject gl9 (glide) ((ins :initform "GL9")))
(defobject gl10 (glide) ((ins :initform "GL10")))

(defobject GFins (m4c-note)
  ((ins :initform "GFins")
   oct.pitch amp func_no attack decay pitch_change)
  (:parameters oct.pitch amp func_no attack decay pitch_change lfrac))

(defobject Buzz (m4c-note)
  ((ins :initform "Buzz")
   oct.pitch amp no_harms ta tdy)
  (:parameters oct.pitch amp no_harms ta tdy lfrac))

(defobject PMins (m4c-note)
  ((ins :initform "PMins")
   oct.pitch amp tac tdc trc susc fcfmratio index tam tdm trm susm)
  (:parameters oct.pitch amp tac tdc trc susc
               fcfmratio index tam tdm trm susm lfrac))

(defobject PMeins (m4c-note)
  ((ins :initform "PMeins")
   oct.pitch amp tac tdc trc susc1 susc2
               fcfmratio index tam tdm trm susm1 susm2)
  (:parameters oct.pitch amp tac tdc trc susc1 susc2
               fcfmratio index tam tdm trm susm1 susm2 lfrac))

(defobject Bpfnois (m4c-note)
  ((ins :initform "Bpfnois")
   oct.pitch amp db_decay bandwidth)
  (:parameters oct.pitch amp db_decay bandwidth lfrac))

(defobject NLF (m4c-note)
  ((ins :initform "NLF")
   oct.pitch amp br ta tdy fc damp dynamic)
  (:parameters oct.pitch amp br ta tdy fc damp dynamic lfrac))

(defobject cornet (NLF) ((ins :initform "cornet")))
(defobject clarinet (NLF) ((ins :initform "clarinet")))
(defobject alt_sax (NLF) ((ins :initform "alt_sax")))
(defobject piano (NLF) ((ins :initform "piano")))

(defobject trumpet (m4c-note)
  ((ins :initform "trumpet")
   oct.pitch ampfac brfac bite ta  tdy  ranamp randfr rancut lfrac)
  (:parameters oct.pitch ampfac brfac bite ta  tdy  ranamp randfr rancut lfrac))


