;;; **********************************************************************
;;; Copyright (c) 89-95, 96 Heinrich Taube.  All rights reserved.
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted and may be copied as long as 
;;; no fees or compensation are charged for use, copying, or accessing
;;; this software and all copies of this software include this copyright
;;; notice.  Suggestions, comments and bug reports are welcome.  Please 
;;; address email to: hkt@zkm.de
;;; **********************************************************************

(in-package :common-music)

(defvar *cm-version* #x100)
(defvar *common-music-directory* )  ; Set by make-cm 
(defvar *cm-site-directory* )       ; Set by make-cm 
(defvar *common-music-package* (find-package :common-music))  

(defun cm-version (&optional (stream t))
  (format stream "Common Music version ~D.~D.~D" 
          (ash (logand *cm-version* #xf00) -8)
          (ash (logand *cm-version* #xf0) -4)
          (logand *cm-version* #xf)))
         
;;;
;;; object status constants
;;;

(defconstant +killed+           (ash 1 0))   ; no output not rescheduled
(defconstant +ending+           (ash 1 1))   ; output not rescheduled
(defconstant +resting+          (ash 1 2))   ; no output
(defconstant +chording+         (ash 1 3))   ; time not increment
(defconstant +unqueued+         (ash 1 4))   ; not rescheduled
(defconstant +hidden+		(ash 1 5))   ; element is skipped
(defconstant +frozen+		(ash 1 6))   ; generator uses thread
(defconstant +dumped+		(ash 1 7))  ; saved bit for archiving
(defconstant +edited+		(ash 1 8))  ; object needs save.
(defconstant +deleted+		(ash 1 9))  ; element is deleted
(defconstant +ephemeral+  	(ash 1 10))  ; unset flushes subelements
(defconstant +system+  		(ash 1 11))  ; element belongs to system
(defconstant +no-link+ 		(ash 1 12))  ; subs sans backpointer
(defconstant +start-unset+      (ash 1 13))  ; flush start time on unset
(defconstant +marked+           (ash 1 14))  ; general marking bit
(defconstant +absolute+         (ash 1 15))  ; time is absolute.
(defconstant +revert-clock+     (ash 1 16))  ; revert system clock.

(defconstant +removed+ (logior +killed+ +unqueued+))
(defconstant +normal-mask+ (logior +resting+ +chording+))

(defconstant +slot-unset+ "unset"
  "Stella's print value for unbound slots.")
(defconstant +unbound-tokens+ 
  '(none unset unbound :none :unset :unbound))

(defparameter *default-output-stream* nil)
(defparameter *default-output-offset* 0.5)

(defvar *tell-stream* *standard-output*)
(defvar *warn-stream* *standard-output*)
(defvar *list-stream* *standard-output*)
(defvar *show-stream* *standard-output*)
(defvar *graph-stream* *standard-output*)

(defun tell-user (&rest args) 
  (apply #'format *tell-stream*   args))

(defun warn-user (&rest args)
  #+ccl (ccl::beep)
  (apply #'format *warn-stream* args))

(defparameter *command-prompting* t 
  "True if Stella should prompt for input.")

(defvar *cm-state* (make-random-state)
  "The random state object for Common Music.")

(defun date-string (&optional stream)
  (multiple-value-bind (second minute hour day month year)
      (get-decoded-time)
    (format stream "~d-~a-~d ~d:~d:~d"
	    day (svref '#("Jan" "Feb" "Mar" "Apr" "May" "Jun" 
			  "Jul" "Aug" "Sep" "Oct" "Nov" "Dec")
		       (1- month))
            (- year 1900) hour minute second)))

(defmacro defprop (sym prop value)
  `(setf (get ',sym ',prop) ,value))

(defun string-append (&rest strings)
  (apply #'concatenate 'string (mapcar #'string strings)))

(defun quotify (token)
  (concatenate 'string "\"" (string token) "\""))

(defun quote-if-necessary (x)
  (if (constantp x) x `(quote ,x)))

(defun objc-string (token)
  (let* ((s (string-downcase (string token)))
         (p (position #\- s)))
    (if p
      (remove #\- (string-capitalize s :start (1+ p)))
      s)))

(defmacro delay-eval (&body body)
  `(function (lambda () ,@body)))

;;;
;;; A constructor and deinitializer for the various part resources
;;;

(defun instance-resource-constructor (resource class)
  (declare (ignore resource))
  (allocate-instance (find-class class)))

(defun instance-resource-deinitializer (resource object)
  (declare (ignore resource))
  (dolist (s (class-slots (class-of object)))
    (slot-makunbound object (slot-definition-name s))))

(defun instance-resource-matcher (resource object class)
  (declare (ignore resource))
  (typep object class))

;;;
;;;
;;;

(defun slots-make-unbound (object &optional slots)
  (dolist (s (or slots (class-slots (class-of object))))
    (slot-makunbound object s)))

(defun compute-slot-and-accessor-pairs (class)
  (let ((pairs '()))
    (flet ((get-accessors (slots)
	     (flet ((slot-and-accessor (slotd)
		      (let ((name (slot-definition-name slotd)))
			(if (not (assoc name pairs))
                          (loop for r in (slot-definition-readers slotd)
			        when (member
                                      r (slot-definition-writers slotd)
                                      :test #'eq :key #'cadr)
			        return (list name r))
			  nil))))
	       (loop with pair for slot in slots
		when (setf pair (slot-and-accessor slot))
		collect pair))))
      (dolist (c (class-precedence-list class))
        (setf pairs (nconc pairs (get-accessors (class-direct-slots c)))))
      pairs)))

(defun find-slot (class name)
  (loop for slot in (class-slots class)
	when (eq (slot-definition-name slot) name)
	  return slot))

(defun class-slot-names (class)
  (mapcar #'slot-definition-name (class-slots class)))

;;;
;;;                         Readtable hacking
;;;

(defvar *common-music-readtable*
	(copy-readtable  #+excl top-level::*readtable*
			 #+lispworks common-lisp-user::*readtable*))

;;;
;;; [] is a read time macro for creating chord streams, and expands into
;;; the CHORD macro call. Pains are taked to insure that the "]" is only
;;; a legal macro terminator if a "[" has actually been typed.
;;;

(defvar *chord-readtable* 
  (copy-readtable #+excl top-level::*readtable*
		  #+lispworks common-lisp-user::*readtable*))

(defun chord-reader (stream char &optional arg)
  (declare (ignore char arg))
  (let ((*readtable* *chord-readtable*)
         members)
    (setf members (read-delimited-list #\] stream t))
    (list* 'chord members)))

;;;
;;; #@ reader for named item streams
;;;

#-mcl
(defun motive-reader (stream subchar arg)
  (declare (ignore subchar arg))
  (values `(motive ,(read stream t nil t))))

#+mcl
(defun motive-reader (stream subchar arg)
  (if (char= (peek-char nil stream t nil t) #\()
      (funcall (get-dispatch-macro-character #\# #\@ 
                                             ccl::%initial-readtable%) 
               stream subchar arg)
    (values `(motive ,(read stream t nil t)))))

;;;
;;; #i reader for item stream definition.
;;;

(defun item-stream-reader (stream subchar arg)
  (declare (ignore subchar arg))
  (let ((p (peek-char nil stream))
        d i)
    (cond ((alpha-char-p p)
           (read-char stream)
           (setf d (peek-char nil stream))
           (cond ((alpha-char-p d)
                  (read-char stream))
                 ((char= d #\() 
                  (setf d #\x))
                 (t (error "Bogus #i character: \"~C\"" d))))
          ((char= p #\()
           (setf p #\c d #\x))
          (t (error "Bogus #i character: \"~C\"" p)))
    (cond ((char-equal p #\c) (setf p 'cycle))
          ((char-equal p #\r) (setf p 'random))
          ((char-equal p #\h) (setf p 'heap))
          ((char-equal p #\g) (setf p 'graph))
          ((char-equal p #\s) (setf p 'sequence))
          ((char-equal p #\a) (setf p 'accumulation))
 	  ((char-equal p #\p) (setf p 'palindrome)) 
          ((char-equal p #\f) (setf p 'function)) 
          ((char-equal p #\o) (setf p 'rotation)) 
          ((char-equal p #\w) (setf p 'rewrite))
          (t (error "\"~C\" not one of C,R,H,G,S,A,P, F or O." p)))
    (cond ((char-equal d #\x) (setf d 'items))
          ((char-equal d #\n) (setf d 'notes))
          ((char-equal d #\r) (setf d 'rhythms))
          ((char-equal d #\p) (setf d 'pitches))
          ((char-equal d #\d) (setf d 'degrees))
          ((char-equal d #\i) (setf d 'intervals))
          ((char-equal d #\s) (setf d 'steps)) 
          ((char-equal d #\a) (setf d 'amplitudes))
	  ((char-equal d #\f) (setf d 'pathnames))
          (t (error "\"~C\" not one of X,N,R,P,D,I,S,A,F." d)))
    (setf i (read stream t nil t))
   `(,d ,@ i in ,p)))
            
(eval-when (load eval)

  ;; [] read macro for creating chord streams
  (set-macro-character #\[ #'chord-reader nil *common-music-readtable*)
  (set-syntax-from-char #\] #\) *chord-readtable* 
  		        #+ExCl top-level::*readtable*
		        #+lispworks system::*std-lisp-readtable*)

  ;; #@ read macro for referencing named item streams.
  (set-dispatch-macro-character #\# #\@ #'motive-reader
                                *common-music-readtable*)

  ;; #i read macro for constructing item streams
  (set-dispatch-macro-character #\# #\i #'item-stream-reader 
                                cm::*common-music-readtable*)

  #+excl (setf top-level::*readtable* *common-music-readtable*)
  #-excl (setf *readtable* *common-music-readtable*)
  )



;;;
;;; most of the commands accept "pair arguments". Pairs are a generalization 
;;; of Lisp's keywords and can be either :keyword or symbols.
;;; values or slot initializations, depending on the context. 
;;;

(defmacro dopairs ((optionvar valuevar pairs) &body body)
  (let ((pvar (gensym)) )
    `(let ((,pvar ,pairs) ,optionvar ,valuevar)
       (loop while ,pvar
             do (setf ,optionvar (pop ,pvar))
                (setf ,valuevar (pop ,pvar))
                ,@body))))

;;; 
;;; canonicalize-pairs normalizes pairs supplied by the user to either slot or
;;; keyword names. it also can perform careful evaluation of values.
;;;

(defun canonicalize-pairs (pairs &optional eval (destruct t) (package :stella))
  (declare (ignore eval))
  (unless (packagep package)
    (setf package (find-package package)))
  (unless destruct (setf pairs (copy-list pairs)))
  (loop for slot = t then (not slot)
        for tail on pairs
        if (and slot (not (eq package (symbol-package (car tail)))))
        do (setf (car tail)
             (let ((name (symbol-name (car tail))))
               (or (find-symbol name package) (intern name package))))
        finally (return pairs)))

(defun pair-value (name pairs &optional default)
  ;(getf pairs name default)
  (let ((name (symbol-name name)))
    (loop for (s v) on pairs by #'cddr
          when (string= name (symbol-name s))
          return v
          finally (return default))))

